

/*********************************************************************
 * (C) Copyright 1999 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *********************************************************************/

/*********************************************************************
 * File: main.c
 * Description: The main routine for the Branch-and-Bound Planner.
 *
 * Author: Joerg Hoffmann 2002
 *  Modified 2004 and 2006 
 *********************************************************************/ 


#include "stdio.h"
#include "stdlib.h"
#include "bb.h"

#include "memory.h"
#include "output.h"

#include "parse.h"
#include "instantiateI.h"
#include "instantiateII.h"
#include "graph.h"
#include "cnfout.h"
#include "cg_encoding.h"
#include "utilities.h"
//Q. Lv 2008-3-27
#include "solver.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#include <pthread.h>
#include <time.h>
#include <pretreat.h>




/*
 *  ----------------------------- GLOBAL VARIABLES ----------------------------
 */

//Q. Lv 2008-3-27
bool gsolution_num;
int  gmini_costs;
int  gmini_costs_layer;
float gmini_costs_time;
int  gcurr_layer;
int  gfirst_soln_costs;
int  gfirst_soln_layer;
float gfirst_soln_time;

double global_time;
bool time_out;
pthread_t thread[10];
Bool	gaction_costs_def = FALSE;
Bool	gmini_costs_require = FALSE;
int		ginit_cost = 0; 
int		gmax_action_cost = 0;

/*******************
 * GENERAL HELPERS *
 *******************/

/* used to time the different stages of the planner
 */
float gtempl_time = 0, greach_time = 0, grelev_time = 0, gconn_time = 0;
float gbuild_time = 0, gsearch_time = 0, gcnf_time = 0;


/* the command line inputs
 */
struct _command_line gcmd_line;

/* number of states that got heuristically evaluated
 */
int gevaluated_states = 0;

/* info on search spaces structure
 */
//int gmax_search_depth = 0;
//int gmax_search_size = 0;



/***********
 * PARSING *
 ***********/

/* used for pddl parsing, flex only allows global variables
 */
int gbracket_count;
char *gproblem_name;

/* The current input line number
 */
int lineno = 1;

/* The current input filename
 */
char *gact_filename;

/* The pddl domain name
 */
char *gdomain_name = NULL;

/* loaded, uninstantiated operators
 */
PlOperator *gloaded_ops = NULL;

/* stores initials as fact_list 
 */
PlNode *gorig_initial_facts = NULL;

/* not yet preprocessed goal facts
 */
PlNode *gorig_goal_facts = NULL;

/* axioms as in UCPOP before being changed to ops
 */
PlOperator *gloaded_axioms = NULL;

/* to store all typed objects 
 */
FactList *gorig_constant_list = NULL;

/* type hierarchy (PDDL) 
 */
type_tree_list gglobal_type_tree_list;

/* helper for types parsing
 */
FactList *gtypes = NULL;

/* the predicates and their types as defined in the domain file
 */
FactList *gpredicates_and_types = NULL;


/*****************
 * INSTANTIATING *
 *****************/

/* global arrays of constant names,
 *               type names (with their constants),
 *               predicate names,
 *               predicate aritys,
 *               defined types of predicate args
 */
Token gconstants[MAX_CONSTANTS];
int gnum_constants = 0;
Token gtype_names[MAX_TYPES];
int gtype_consts[MAX_TYPES][MAX_TYPE];
Bool gis_member[MAX_CONSTANTS][MAX_TYPES];
int gtype_size[MAX_TYPES];
int gnum_types = 0;
Token gpredicates[MAX_PREDICATES];
int garity[MAX_PREDICATES];
int gpredicates_args_type[MAX_PREDICATES][MAX_ARITY];
int gnum_predicates = 0;





/* the domain in integer (Fact) representation
 */
Operator_pointer goperators[MAX_OPERATORS];
int gnum_operators;
Fact gfull_initial[MAX_INITIAL];
int gnum_full_initial = 0;
Fact ggoal[MAX_STATE];
int gnum_goal = 0;






/* stores inertia - information: is any occurence of the predicate
 * added / deleted in the uninstantiated ops ?
 */
Bool gis_added[MAX_PREDICATES];
Bool gis_deleted[MAX_PREDICATES];

/* splitted initial state:
 * initial non static facts,
 * initial inertia facts of arity > 1
 */
Fact ginitial[MAX_STATE];
int gnum_initial = 0;
Fact ginertia[MAX_INITIAL];
int gnum_inertia = 0;

/* the type numbers corresponding to any unary inertia
 */
int gtype_to_predicate[MAX_PREDICATES];
int gpredicate_to_type[MAX_TYPES];

/* (ordered) numbers of types that new type is intersection of
 */
TypeArray gintersected_types[MAX_TYPES];
int gnum_intersected_types[MAX_TYPES];





/* intermediate step: store minimal action info, use it
 * to determine relevant facts, number of actions and final form
 * before creating actual actions
 */
ActionTemplate *gtemplates = NULL;
int gnum_templates = 0;




/* store the final "relevant facts"
 */
Fact grelevant_facts[MAX_RELEVANT_FACTS];
int gnum_relevant_facts = 0;
int gnum_pp_facts;


/* the (fully instantiated) domain in integer 
 * (number of relevant Fact) representation
 */
Action *gactions = NULL;
int gnum_actions;
State ginitial_state;
State ggoal_state;

/**********************
 * CONNECTIVITY GRAPH *
 **********************/

/* one fact array, ..
 */
FtConn *gft_conn;
int gnum_ft_conn;

/* one ops (actions) array
 */
OpConn *gop_conn;
int gnum_op_conn;



int gcword_size; 
int gnum_ft_bit;
int gnum_op_bit_at[MAX_GRAPH];
int gnum_ops = 0, gmax_block = 0;

/* the goals in bit-format...
 */
BitVector *gbit_goal_state;



/* some technical graph... things
 */





/* the ops not yet in the graph
 */
IntList *gout_ops = NULL;

/* global op-state resp. ft-state lists
 */ 
IntList *gin_ops = NULL;
IntList *gin_prev_ops = NULL;
IntList *gin_fts = NULL;
IntList *gin_prev_fts = NULL;

/* we want to have fast access, during constraint propagation,
 * to the fts that can (maximally) be present at the layer;
 * can easily store this by means of pointers into the gin_fts list
 * which we got anyway.
 */
IntList *gin_fts_at[MAX_GRAPH];
/* the present ops are good to have for enforcing a positive constraint:
 * we must then negatively enforce the exclusive ops. don't wanna store
 * all the exclusives in a list but at least we can skip those ops that
 * aren't even there. means as above.
 */
IntList *gin_ops_at[MAX_GRAPH];

/* current mutexes: exclusives speedup
 */
IntPair *gin_op_mutex_pairs = NULL;
IntPair *gin_ft_mutex_pairs = NULL;

/* information about current state of graph, needed for level off test
 */
int gin_ft_count = 0, gin_ft_exclusion_count = 0;
int gin_op_count = 0, gin_op_exclusion_count = 0;

/* is TRUE iff graph has levelled off.
 */
Bool gsame_as_prev_flag = FALSE;

/* stores the time step at which graph has levelled off.
 */
int gfirst_full_time = 0;


struct tms gstart, gend;
struct tms start, end;

/*
 *  ----------------------------- HEADERS FOR PARSING ----------------------------
 * ( fns defined in the scan-* files )
 */

void get_fct_file_name( char *filename );
void load_ops_file( char *filename );
void load_fct_file( char *filename );

/*
 *  ----------------------------- MAIN ROUTINE ----------------------------
 */

/* start - globals and prototypes added by sjn */
const int SAT = 0;
const int UNSAT = 2;
int get_solution(solver* s, FILE* fp, FILE* solnFile);
int create_solution(solver* s);
/* end */






int main( int argc, char *argv[] )
{
	srand((unsigned)time(NULL));
	times( &gstart );
	printf("MaxPlanII beta.\n");
	long solver_start_clk = (long)clock();
	// resulting name for ops file
	
	char ops_file[MAX_LENGTH] = "";
	char fct_file[MAX_LENGTH] = "";
  
	gcmd_line.cnflayer = -1;
	gcmd_line.cnfout = -1;
	gcmd_line.solverOut = FALSE;
	gcmd_line.prune = FALSE;
	gcmd_line.binary_clause_only = 0;
	strcpy(gcmd_line.cnfFileName, "CNF");
	gcmd_line.makeCNF = 0;
	gcmd_line.londex_mode = 0; 
	
	// command line treatment
   
	if ( argc == 1 || ( argc == 2 && *++argv[0] == '?' ) || strcmp(argv[1],"-h")==0 ) {
		bb_usage();
		exit( 1 );
	}
	if ( !process_command_line( argc, argv ) ) {
		printf("\n\nEXIT: bad arguments detected\n\n");
		bb_usage();    
		exit( 1 );
	}



	/* make file names
	*/

	/* one input name missing
	*/
	if ( !gcmd_line.ops_file_name || 
			!gcmd_line.fct_file_name ) {
		fprintf(stdout, "\nbb: two input files needed\n\n");
		bb_usage();      
		exit( 1 );
	}
	/* add path info, complete file names will be stored in
	* ops_file and fct_file 
	*/
	
	
	sprintf(ops_file, "%s%s", gcmd_line.path, gcmd_line.ops_file_name);
	sprintf(fct_file, "%s%s", gcmd_line.path, gcmd_line.fct_file_name);
	
	char cg_domain[MAX_LENGTH] = "";
	char cg_problem[MAX_LENGTH] = "";
	
	sprintf(cg_domain,"%s%s",gcmd_line.path,gcmd_line.cg_domain_file_name);
	sprintf(cg_problem,"%s%s",gcmd_line.path,gcmd_line.cg_problem_file_name);
  
	//pretreat domain and problem file, produce the file without cost information   
	char preops_file[MAX_LENGTH] ="preops_file";
	char prefct_file[MAX_LENGTH] ="prefct_file";

	pretreat_ops_file(ops_file, preops_file);
	pretreat_fct_file(fct_file, prefct_file);


	/* parse the input files
	*/

	/* start parse & instantiation timing
	*/
	times( &start );
	/* domain file (ops)
	*/
	if ( gcmd_line.display_info >= 1 ) {
		printf("\nbb: parsing domain file");
	} 
	/* it is important for the pddl language to define the domain before 
	* reading the problem 
	*/
	load_ops_file( preops_file );
	/* problem file (facts)
	*/  
	if ( gcmd_line.display_info >= 1 ) {
		printf(" ... done.\nbb: parsing problem file"); 
	}
	
	load_fct_file( prefct_file );
	
	if ( gcmd_line.display_info >= 1 ) {
		printf(" ... done.\n\n");
	}

	/* This is needed to get all types.
	*/
	build_orig_constant_list();

	/* last step of parsing: see if it's a STRIPS domain!
	*/
	if ( !make_strips_domain() ) {
		printf("\n\nEXIT: bb: this is not a STRIPS problem! can't be handled by this version.\n\n");
		exit( 1 );
	}

	/* now instantiate operators; 
	*/

	/* start by collecting all strings and thereby encoding 
	* the domain in integers.
	*/
	encode_domain_in_integers();

	/* inertia preprocessing:
   *   - collect inertia information
   *   - split initial state into
   *        - types for unary inertia
   *        - array containing all inertia relations
   *        - array containing non - static relations
   *   - encode unary inertia in op preconds into types
   */
	do_inertia_preprocessing();

  /* unify inertia preconds with initial state,
   * then multiply remaining parameters;
   * create one ActionTemplate for each possible parameter combination
   *
   * --- the actual instantiation part, creating all possible instances ---
   */
	build_action_templates();

	times(&end);
	TIME( gtempl_time );

	times(&start);

  /* perform a simple reachabilty analysis
   * (added facts fixpoint)
   * to find out which facts can (at most) be made true in domain
   */
	perform_reachability_analysis();

	times(&end);
	TIME( greach_time );

	times(&start);

  /* as a last step, collect relevant facts (deleted ini's and
   * added non ini's) and remove others from domain
   */
	collect_relevant_facts();

	times(&end);
	TIME( grelev_time );

	times(&start);
	//if the domain defien action-costs, pretreat the file for londex
	if(gaction_costs_def == TRUE && gmini_costs_require == TRUE){
		printf("\naction-costs defined...\n");fflush(stdout);
		scan_ops_file(ops_file);
		scan_fct_file(fct_file);
		
		char cg_preops_file[MAX_LENGTH] = "cg_preops_file";
		char cg_prefct_file[MAX_LENGTH] = "cg_prefct_file";
		pretreat_ops_file(cg_domain, cg_preops_file);
		pretreat_fct_file(cg_problem, cg_prefct_file);

		strncpy(gcmd_line.cg_domain_file_name, cg_preops_file,MAX_LENGTH);
		strncpy(gcmd_line.cg_problem_file_name, cg_prefct_file,MAX_LENGTH);
	}
	else{
		for( int i = 0; i < gnum_operators; i++ ){
			goperators[i]->cost = 1;
		}
		strncpy( gcmd_line.cg_domain_file_name, cg_domain, MAX_LENGTH);
		strncpy( gcmd_line.cg_problem_file_name, cg_problem, MAX_LENGTH);
	}
	
  	/* now build globally accessible connectivity graph
   	*/
	printf("before Scan CG file...  ");
	build_connectivity_graph();

	printf("Scan CG file...  ");
  	scan_cg_file();
  /* insert bit vectors for fast lookup
   *
   * (NOTE: bit vector sizes are static across graph here;
   *  (in difference to IPP) -- otherwise pre-computation 
   *  not possible as done in these parts)
   */
  	insert_bit_vectors();
  	times(&end);
  	TIME( gconn_time );
  
  	int maxLevel = 100,layer;
	gsolution_num = 0;
	time_out = false;
	gfirst_soln_costs = MAX_COST;
	gfirst_soln_layer = maxLevel;
	gfirst_soln_time = gcmd_line.tlimit;
	gmini_costs = MAX_COST;
	gmini_costs_layer = maxLevel;
	gmini_costs_time = gcmd_line.tlimit;
	gcmd_line.makeCNF = 0;
	for( layer = 1 ; layer<maxLevel ;layer++){
		
		if(time_out == true){
			break;
		}
		
		printf( "\n******  Solve with %d layers Start******\n", layer );
		gcmd_line.cnflayer = layer;
	    gcurr_layer = layer;	

		execute_solver();
		if(gsolution_num > 0)
			break;
	}

	//printf("begin solve in multi-thread...\n\n");fflush(stdout);
	/**************  thread join  ******************/	
#if(!DEBUG)
	if(gmini_costs_require == TRUE){
		if(pthread_join(thread[0],NULL)){
			perror("cannot join thread 0");
			return 1;
		}
	
		if(pthread_join(thread[1],NULL)){
			perror("cannot join thread 1");
			return 1;
		}
	
		if(pthread_join(thread[2],NULL)){
			perror("cannot join thread 2");
			return 1;
		}
/*
		if(pthread_join(thread[3],NULL)){
			perror("cannot join thread 3");
			return 1;
		}

	if(pthread_join(thread[4],NULL)){
		perror("cannot join thread 4");
		return 1;
	}
	
	if(pthread_join(thread[5],NULL)){
		perror("cannot join thread 5");
		return 1;
	}

	if(pthread_join(thread[6],NULL)){
		perror("cannot join thread 6");
		return 1;
	}
	if(pthread_join(thread[7],NULL)){
		perror("cannot join thread 7");
		return 1;
	}
	if(pthread_join(thread[8],NULL)){
		perror("cannot join thread 8");
		return 1;
	}
	
	if(pthread_join(thread[9],NULL)){
		perror("cannot join thread 9");
		return 1;
	}*/
	}
#endif
	check_memory_usage();
	if(time_out == true){
		if(gsolution_num > 0){
			printf("Solver runs with time out.\n");
			output_planner_info();
		}
		else{
			printf("Solver runs with time out.(%d seconds)\n Failed to find a solution.\n", gcmd_line.tlimit );
		}
	}else{
		output_planner_info();
	}

	free_memory();
	return 1;
}


/*
 *  ----------------------------- HELPING FUNCTIONS ----------------------------
 */
void thread_create()
{
#if(!DEBUG)
	if(pthread_create(&thread[0],NULL,thread_solve,NULL)){
		perror("cannot create thread 1");
		return 1;
	}

	if(pthread_create(&thread[1],NULL,thread_solve,NULL)){
		perror("cannot create thread 2");
		return 1;
	}	
	
	if(pthread_create(&thread[2],NULL,thread_solve,NULL)){
		perror("cannot create thread 3");
		return 1;
	}
/*
	if(pthread_create(&thread[3],NULL,thread_solve,NULL)){
		perror("cannot create thread 4");
		return 1;
	}
	
	if(pthread_create(&thread[4],NULL,thread_solve,NULL)){
		perror("cannot create thread 5");
		return 1;
	}

	if(pthread_create(&thread[5],NULL,thread_solve,NULL)){
		perror("cannot create thread 6");
		return 1;
	}	
	
	if(pthread_create(&thread[6],NULL,thread_solve,NULL)){
		perror("cannot create thread 6");
		return 1;
	}

	if(pthread_create(&thread[7],NULL,thread_solve,NULL)){
		perror("cannot create thread 7");
		return 1;
	}

	if(pthread_create(&thread[8],NULL,thread_solve,NULL)){
		perror("cannot create thread 8");
		return 1;
	}

	if(pthread_create(&thread[9],NULL,thread_solve,NULL)){
		perror("cannot create thread 9");
		return 1;
	}	
*/

#endif
}

void free_memory()
{
	int i;
	for( i = 0; i < gnum_constants; i++){
		free(gconstants[i]);
	}
	for( i = 0; i < gnum_types; i++){
		free(gtype_names[i]);
	}
	for( i = 0; i < gnum_predicates; i++)
		free(gpredicates[i]);
	for( i = 0; i < gnum_operators; i++){
		if(goperators[i]!=NULL)
			free_Operator(goperators[i]);
	}
	ActionTemplate *tmp;
	while(gtemplates!=NULL){
		 tmp = gtemplates;
		 gtemplates = gtemplates->next;
		 free(tmp);
	}
	Action *tmp1;
	while(gactions!=NULL){
		tmp1 = gactions;
		gactions = gactions->next;
		free(tmp1);
	}
	if(gft_conn){
		for( i = 0; i < gnum_ft_conn; i++){
			if(gft_conn[i].A){
				free(gft_conn[i].A);
				free(gft_conn[i].D);
				free(gft_conn[i].P);
				if(gft_conn[i].bit_Mux)
					free(gft_conn[i].bit_Mux);
			}
			for( int j = 0; j < MAX_GRAPH+1; j++){
				free_FtLevelInfo(gft_conn[i].info_at[j]);
			}
		}
		free(gft_conn);
	}
	if(gop_conn){
		for( i = 0; i < gnum_op_conn; i++){
			if(gop_conn[i].A){
				free(gop_conn[i].A);
				free(gop_conn[i].D);
				free(gop_conn[i].P);
				if(gop_conn[i].bit_Mux)
					free(gop_conn[i].bit_Mux);
				free(gop_conn[i].bit_A);
				free(gop_conn[i].bit_D);
			}
			for( int j = 0; j < MAX_GRAPH; j++)
				free_OpLevelInfo(gop_conn[i].info_at[j]);
		}
		free(gop_conn);
	}
	char args[1000];
	sprintf(args, "rm -f CG_Temp preops_file prefct_file cg_preops_file cg_prefct_file");
	system(args);

}


void output_planner_info( void )
{
	printf("\nInfo:   %7.2lf seconds global time", get_global_time());
	printf("\n        %7.2lf seconds total encoding time", encoding_time);
	printf("\n        %7.2lf seconds total SAT solving time", solving_time);
	printf("\n        %7.2f MB memory used",max_memory_usage);	
	
	printf("\nOther:");
	printf("\n        %ld londex constraints",londex_count);
	printf("\n        %ld mutex constraints",mutex_count);
	printf("\n        %ld overlapping constraints",long_in_mutex_count);
	printf("\n");
	
	if(gsolution_num > 0){
		printf("\nFirst solution info:");
		printf("\n	layer %d",gfirst_soln_layer);
		printf("\n	solve time %7.2lf",gfirst_soln_time);
		printf("\n	costs %d",gfirst_soln_costs); 
		
		if(gmini_costs_require == TRUE){
			printf("\nMinimal solution info:");
			printf("\n	layer %d",gmini_costs_layer);
			printf("\n	solve time %7.2lf",gmini_costs_time);
			printf("\n	costs %d",gmini_costs);
			printf("\n\n	have found %d solutions",gsolution_num);
		}
	}
	else
		printf("\nCan't find a valid solution!");
	printf("\n\n");fflush(stdout);
}

void output_planner_info_file(FILE* fp, int numSteps) {
    
    fprintf(fp, "; Time %7.2lf\n", get_global_time());
    fprintf(fp, "; MakeSpan %d  current layer %d\n", numSteps,gcurr_layer);
	if(gaction_costs_def==TRUE&&gmini_costs_require==TRUE){
		fprintf(fp, "; initialize total-cost to %d\n",ginit_cost); 
		fprintf(fp, "; Minimal total-costs %d found at %d layer\n",gmini_costs,gmini_costs_layer);
	}
}

void bb_usage( void ){

  printf("\nusage of Maxplan II:\n");

  printf("\nOPTIONS   DESCRIPTIONS\n\n");
  printf("-p -path <str>    path for operator and fact file\n");
  printf("-o -domain <str>    operator strips file name\n");
  printf("-f -problem <str>    fact strips file name\n\n");
  
  printf("-cgdomain <str>    operator ADL/strips file name for producing Causal Graph\n");
  printf("-cgproblem <str>    fact ADL/strips file name for producing Causal Grapth\n\n");
  printf("-l  <num>   goal layer for CNF\n\n");

  printf("-G <0 or 1> (0) create CNF output or (1) build final solution\n");
  printf("-b <str>    CNF output file name\n");
  printf("-t <0 or 1> (1) CNF output includes only unary/binary clauses - others ignored\n");
  printf("-S <str>    Input Solution File Name (only when -G 1 is used)\n");
  printf("-F -out <str>    Final Output Solution File Name (only when -G 1 is used)\n");
  printf("-V <str>    Variables File Name - list all variables (only when -G 1 is used)\n");
  printf("-londexm <0 or 1>  (0) just use mutex, (1) use londex\n");
  printf("-C          CNF formula output (preset: %d); at layer <-l>\n",
	 gcmd_line.cnfout);
  printf("      0     none\n");
  printf("      1     action-based\n");
  printf("      2     gp-style action-based\n");
  printf("      3     gp-based\n");
  printf("      4     thin gp-based\n\n");




  return;

  printf("-i <num>    run-time information level( preset: 1 )\n");
  printf("      0     only times\n");
  printf("      1     problem name, planning process infos\n");
  printf("    101     cleaned up STRIPS problem\n");
  printf("    102     collected string tables\n");
  printf("    103     encoded domain\n");
  printf("    104     predicates inertia info\n");
  printf("    105     splitted initial state\n");
  printf("    106     extended types table\n");
  printf("    107     all action templates\n");
  printf("    108     inertia cleaned up operator schemata\n");
  printf("    109     reachability analysis results\n");
  printf("    110     final domain representation\n");
  printf("    111     connectivity graph\n\n");
  printf("    112     bits in connectivity graph\n\n");
  
  printf("\n-d <num>    switch on debugging\n\n");

}


/* begin additions by sjn */
/*
 * fixed solving_result name by R.H.bb
 */
 
int create_solution(solver* s) {

    FILE* fp;
    FILE* finalSoln;
    int result;

    /* open files for input/output */
    if ( (fp = fopen( "bb_solving_result_cnf", "r" )) == NULL ) {
        printf("\n\nEXIT: can't open resulting input solution file.\n\n");
 		exit(1);
    }
	
    if ( (finalSoln = fopen(gcmd_line.final_solution, "w") ) == NULL ) {
        if ( EOF == fclose(fp) )
            printf("\nExit: unable to close final solution file --> resource leak\n");
        printf("\n\nEXIT: can't open file to write final solution.\n\n");
        exit(1);
    }
	
    /* write final solution to external file */
    result = get_solution(s,fp, finalSoln);
	if ( EOF == fclose(fp) ) {
        if ( EOF == fclose(finalSoln) )
            printf("\n\nEXIT: unable to close final solution file --> resource leak\n\n");
        printf("\n\nEXIT: unable to close input solution file --> resource leak\n\n");
        exit(1);
    }

    if ( EOF == fclose(finalSoln) ) {
        printf("\n\nExit: unable to close final solution file --> resource leak\n\n");
        exit(1);
    }
    return(result);
}

typedef struct _PairVals { /* simple structure to help with sorting/pruning final solution */
    int first;
    int second;
    int keep_state; /* if > 1 --> keep in final solution, else throw away */
} PairVals;

void sort_solution(PairVals* pv, int size) {
    const int done = 0;
    int pre = 0, curr = 1, max = -1;
    PairVals tmp;

    if ( curr >= size )
        return;

    /* pv should be almost (if not completely) sorted already --> 
       this algorithm isn't too bad... */
    while ( ! done ) {
        if ( pv[curr].first < pv[pre].first ) { /* swap */
            tmp.first = pv[curr].first;
            tmp.second = pv[curr].second;
            tmp.keep_state = pv[curr].keep_state;
            pv[curr].first = pv[pre].first;
            pv[curr].second = pv[pre].second;
            pv[curr].keep_state = pv[pre].keep_state;
            pv[pre].first = tmp.first;
            pv[pre].second = tmp.second;
            pv[pre].keep_state = tmp.keep_state;
            if ( pv[pre].first < max ) { /* decrement */
                if ( pre != 0 ) {
                    pre -= 1; 
                    curr -= 1;
                    continue;
                }
            }
        }
        else
            max = pv[curr].first;
        if ( ++curr >= size )
            return;
        ++pre;
    } /* while */
}

void removeArr(int* arr, int which, int size) {
    if ( which == size - 1 ) {
        arr[which] = 0;
        return;
    }

    while ( which + 1 < size ) {
        arr[which] = arr[which+1];
        if ( ++which == size - 1 ) 
            arr[which] = 0;
    } /* moving elements to lower positions of arr */    
}

void prune(solver* s, PairVals* varList, int num_vars) { /* varList has been sorted on plan-levels */
    int i = 0, j = 0, k = 0, m = 0, currentValue = -1, n = 0, p = 0, breakAgain = 0;
    int ok = 1, mark = 0, ft;
    int conditions[10000];
	int*	lop_to_code = s->lop_to_code;
	int*	ltime_to_code = s->ltime_to_code;

    /* Look at initial facts; only keep actions at level 0 that can result from initial conditions */
    /* This was a hack to get plans using thin-gp-based and gp-based to work; need to find root cause still */
    /* I believe the problem is how no-ops are handled in these other encoding approaches */
    for ( k = 0; k < ginitial_state.num_F; ++k )
        conditions[++currentValue] = ginitial_state.F[k];
    i = -1;
    while ( ++i < num_vars && varList[i].first == 0 ) {
        k = 0;
        for ( n = 0; n < gop_conn[lop_to_code[varList[i].second]].num_P; ++n ) {
            for ( p = 0; p <= currentValue; ++p ) {
                if ( gop_conn[lop_to_code[varList[i].second]].P[n] == conditions[p] ) {
                    ++k;
                    break;
                }
            }
        }
        if ( k != gop_conn[lop_to_code[varList[i].second]].num_P )
            varList[i].keep_state = -1;
     }
    currentValue = -1;

    /* Search forward for steps that are unnecessary at level 0 */
    i = -1;
    while ( ++i < num_vars && varList[i].first == 0 ) {
        if ( varList[i].keep_state < 0 )
            continue;
        k = 0;
        for ( n = 0; n < gop_conn[lop_to_code[varList[i].second]].num_A; ++n ) {	    
            for ( m = 0; m <= currentValue; ++m ) {
	        if ( gop_conn[lop_to_code[varList[i].second]].A[n] == conditions[m] )
		    break; /* nothing new */
            }
            if ( m > currentValue ) {
              k = 1;
	      conditions[++currentValue] = gop_conn[lop_to_code[varList[i].second]].A[n];
            }
        }
        if ( k == 0 )
	  varList[i].keep_state = -1;
    }

    /* Iterate by level; updating conditions as steps are determined necessary; remove unnecessaries */
    i = -1; j = -1;
    while ( ++i < num_vars - 1 ) {
      while ( ++j < num_vars && varList[j].first == varList[i].first ) ;

      mark = j;
      while ( j < num_vars && varList[mark].first == varList[j].first ) {
        if ( varList[j].keep_state < 0 ) {
            ++j;
            continue;
        }
        k = 0;
        for ( n = 0; n < gop_conn[lop_to_code[varList[j].second]].num_A; ++n ) {
	  for ( m = 0; m <= currentValue; ++m ) {
	    if ( gop_conn[lop_to_code[varList[j].second]].A[n] == conditions[m] )
	      break;
          }
          if ( m > currentValue ) {
	    k = 1;
            conditions[++currentValue] = gop_conn[lop_to_code[varList[j].second]].A[n];
          }
        }
        if ( k == 1 ) { /* keep */
	  for ( n = 0; n < gop_conn[lop_to_code[varList[j].second]].num_D; ++n ) {
            for ( m = 0; m <= currentValue; ++m ) {
              if ( gop_conn[lop_to_code[varList[j].second]].D[n] == conditions[m] ) {
                removeArr(conditions, m, currentValue + 1);
                --currentValue;
	        break;
              }
            }
          }
        }
        else
          varList[j].keep_state = -1;
        ++j;
      } /* while */
      i = mark-1;
      --j;
    } /* while */

    currentValue = -1;
    i = 0; j = 0; m = 0; n = 0;
    
    /* load up all goal conditions */
    for ( i = 0; i < ggoal_state.num_F; i++ )
        conditions[++currentValue] = ggoal_state.F[i];

    /* mark ops in final layer which contribute to the goals */
    i = 0; j = 0;
    for ( i = num_vars - 1; i >= 0; --i ) {
        if ( varList[i].first != varList[num_vars - 1].first )
            break;
        else if ( varList[i].keep_state < 0 )
	    continue;

        for ( m = 0; m <= currentValue; ++m ) {
            for ( n = 0; n < gop_conn[lop_to_code[varList[i].second]].num_A; ++n ) {
                if ( gop_conn[lop_to_code[varList[i].second]].A[n] == conditions[m] ) {
                    /* match found */
                    varList[i].keep_state = 2;
                    removeArr(conditions, m, currentValue + 1);
                    --currentValue;
                    m -= 1;
                    break;
                }
            } /* for */
        } /* for */
    } /* for */

    /* collect union (unique set) of all preconditions at plan level i that truly contribute to plan */
    for ( i = num_vars - 1; i >= 0; ) {
        j = i;
        while ( (--j >= 0) && (varList[j].first == varList[i].first) ) ;


        while ( i > j ) { 
            if ( varList[i].keep_state > 1 ) { /* more preconditions to add */
                for ( k = 0; k < gop_conn[lop_to_code[varList[i].second]].num_P; ++k ) {
                    for ( p = 0; p <= currentValue; ++p ) {
                        if ( gop_conn[lop_to_code[varList[i].second]].P[k] == conditions[p] ) {
                            ok = 0; /* do not add --> already in the current list */
                            break;
                        }
                    }
                    if ( ok )
                        conditions[++currentValue] = gop_conn[lop_to_code[varList[i].second]].P[k];                    
                    ok = 1;
                }                    
            }
            --i;
        } /* while */
        if ( j >= 0 ) {
            while ( (--j >= 0) && (varList[j].first == varList[i].first) ) ;
        }

        /* look at preceding plan level and see which actions stay and which must go */
        for ( k = j + 1; k <= i && (k > -1); ++k ) {
            if ( varList[k].keep_state < 0 )
                continue;
            varList[k].keep_state = 0;
            breakAgain = 0;
            for ( m = 0; m <= currentValue; ++m ) {
                for ( n = 0; n < gop_conn[lop_to_code[varList[k].second]].num_A; ++n ) {
                    if ( gop_conn[lop_to_code[varList[k].second]].A[n] == conditions[m] ) {
                        /* match found */
                        varList[k].keep_state = 2;
                        removeArr(conditions, m, currentValue + 1);
                        if ( --currentValue < 0 ) breakAgain = 1;
                        m -= 1;
                        break;
                    }
                } /* for */
                if ( breakAgain ) break;
            } /* for */
            if ( breakAgain ) break;
        } /* for */
    } /* for */


    /* look at initial state --> finish off all necessary conditions */
    for ( m = 0; m <= currentValue; ++m ) {
        for ( k = 0; k < ginitial_state.num_F; ++k ) {
            ft = ginitial_state.F[k];
            if ( ft == conditions[m] ) {
                removeArr(conditions, m, currentValue + 1);
                if ( --currentValue < 0 )
                    return;
                m -= 1;
                break;
            }
        }
    }

    /* Ensure all preconditions of all ops kept in final solution have been satisfied */
	if ( currentValue != -1 ) {
        printf("\n\nEXIT : prune() left some preconditions open.. %d preconditions left over\n\n", ++currentValue);
        printf("NOT A SOLUTION\n\n");
		exit(UNSAT);
     }
	return(SAT);
}

int get_solution(solver* s, FILE* fp, FILE* solnFile) {
    /* locals */
    long loc = 1, markEnd = -1, beg = 0;
    char c;
    long done = 0, value = 0, idx = -1, first = 1;
    char toConvert[20];
    int i = 0, j = 0;
    int counter = 0;
    int maxCounter = MAX_CNF_VARS / 2 + 1;
    int numSteps = -1;
    static PairVals toOrder[MAX_CNF_VARS / 2 + 1];    
	//Q. Lv 2008-5-9
	int*	lop_to_code = s->lop_to_code;
	int*	ltime_to_code = s->ltime_to_code;

    /* get end of file position */
    fseek(fp, beg, SEEK_END);
    loc = ftell(fp);

    while ( loc > beg ) {
        fseek(fp, loc, SEEK_SET);
        c = fgetc(fp);
        if ( isdigit(c) ) {
            markEnd = loc + 1;
            break;
        }
        --loc;
    }
    if ( loc <= beg ) {
        printf("\n\nEXIT: Improper solution format.\n\n");
        fclose(fp);
        fclose(solnFile);
        exit(1);
    }

    while ( beg < markEnd ) {        
        fseek(fp, beg, SEEK_SET);
        c = fgetc(fp);
        if ( isdigit(c) || (c == '-') ) /* # or -sign of # */
            break;
        ++beg;  
    }
    if ( beg >= markEnd ) {
        printf("\n\nEXIT: Improper solution format.\n\n");
        fclose(fp);
        fclose(solnFile);
        exit(1);
    }

    /* grab solution to CNF sentence */
    fseek(fp, beg, SEEK_SET);

    while ( beg <= markEnd ) {

        c = fgetc(fp);
        switch(c) {
            case ' ': case '\n':
                if ( first ) { /* just front garbage */
                    fseek(fp, ++beg, SEEK_SET);
                    continue;
                }
                if ( ! done ) {
                	
                    toConvert[++idx] = '\0';
                    value = atol(toConvert);
                    
                    if ( value > 0 && (gop_conn[lop_to_code[value]].noop_for == -1) ) {
                        /* found part of final solution */
                        toOrder[counter].first = ltime_to_code[value]; /* sort on this */
                        toOrder[counter].second = value;
		        		toOrder[counter].keep_state = 0;
                        if ( ltime_to_code[value] > numSteps )
                            numSteps = ltime_to_code[value];
                        if ( ++counter > maxCounter ) {
                            printf("\n\nEXIT: Too many variables asserted????\n\n");
                            fclose(fp);
                            fclose(solnFile);
                            exit(1);
                        }     
                    } /* if */                    
                    idx = -1;
                    done = 1;
                } /* if */
                break;
            default:
                if ( ++idx > 19 ) {
                    printf("\n\nEXIT: wow, that's a big number - too big\n\n");
                    fclose(fp);
                    fclose(solnFile);
                    exit(1);
                }
                toConvert[idx] = c;  
                done = 0;
                first = 0;
        }; /* switch */
        fseek(fp, ++beg, SEEK_SET);
    } /* while */

    /* sort final solution */
    sort_solution(toOrder, counter);

    /* prune away actions that don't contribute */
   prune(s, toOrder, counter);
	
    /* send final solution to file; print out any pruned steps */
    output_planner_info_file(solnFile, ++numSteps);
    for ( j = 0; j < counter; ++j ) {
      if (toOrder[j].keep_state > 1) {
        value = toOrder[j].second;
		int op_index = gop_conn[lop_to_code[value]].op;
        fprintf(solnFile, "%d: (", ltime_to_code[value]);
        if ( goperators[op_index]->name )
	        fprintf(solnFile, "%s", goperators[op_index]->name);                    	
        for ( i=0; i<goperators[op_index]->num_vars; i++ )
		    fprintf(solnFile, " %s", gconstants[gop_conn[lop_to_code[value]].inst_table[i]]);
        fprintf(solnFile, ") [1]\n"); 
      } else {
	    //printf("Pruned:  ");
        value = toOrder[j].second;
        //printf("%d: (", ltime_to_code[value]);
        /*
        if ( goperators[gop_conn[lop_to_code[value]].op]->name )
	      printf("%s", goperators[gop_conn[lop_to_code[value]].op]->name);                    	
        for ( i=0; i<goperators[gop_conn[lop_to_code[value]].op]->num_vars; i++ )
	      printf(" %s", gconstants[gop_conn[lop_to_code[value]].inst_table[i]]);
        //printf(") [1]\n");
         * */ 
      }
    }
    return(SAT);
}
/* end additions by sjn */
